//
// +---------------------------------------------------------------------------+
// | Blip.fm Playlist Controller v2.0                                          |
// +---------------------------------------------------------------------------+
// | Copyright (c) 2008 Fuzz Artists, Inc.                                     |
// | All rights reserved.                                                      |
// |                                                                           |
// | Redistribution and use in source and binary forms, with or without        |
// | modification, are permitted provided that the following conditions        |
// | are met:                                                                  |
// |                                                                           |
// | 1. Redistributions of source code must retain the above copyright         |
// |    notice, this list of conditions and the following disclaimer.          |
// | 2. Redistributions in binary form must reproduce the above copyright      |
// |    notice, this list of conditions and the following disclaimer in the    |
// |    documentation and/or other materials provided with the distribution.   |
// |                                                                           |
// | THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR      |
// | IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES |
// | OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.   |
// | IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,          |
// | INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  |
// | NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, |
// | DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY     |
// | THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT       |
// | (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF  |
// | THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.         |
// +---------------------------------------------------------------------------+
// | For help with this library, contact api@blip.fm                           |
// +---------------------------------------------------------------------------+
//
//  Requirements:
//      YUI (http://developer.yahoo.com/yui)
//          - DOM
//          - Event
//          - Connection Manager
//          - JSON
//

// Alias $ to YUI
var $  = YAHOO.util.Dom.get;
var $E = YAHOO.util.Event;
var $D = YAHOO.util.Dom;

/**************************************
* Linked List
***************************************/

// Blip playlist is a linked list
function LinkedList() {}
LinkedList.prototype = {
    list: new Array(),
    length: 0,
    first: null,
    last: null
};
LinkedList.Circular = function() {};
LinkedList.Circular.prototype = new LinkedList();
LinkedList.Circular.prototype.append = function(node) {
    if (this.first === null) {
        node.prev = node;
        node.next = node;
        this.first = node;
        this.last = node;
    } else {
        node.prev = this.last;
        node.next = this.first;
        this.first.prev = node;
        this.last.next = node;
        this.last = node;
    }
    this.length++;
};
LinkedList.Circular.prototype.insertAfter = function(node, newNode) {
    newNode.prev = node;
    newNode.next = node.next;
    node.next.prev = newNode;
    node.next = newNode;
    if (newNode.prev == this.last) { this.last = newNode; }
    this.length++;
};
LinkedList.Circular.prototype.remove = function(node) {
    if (this.length > 1) {
        node.prev.next = node.next;
        node.next.prev = node.prev;
        if (node == this.first) { this.first = node.next; }
        if (node == this.last) { this.last = node.prev; }
    } else {
        this.first = null;
        this.last = null;
    }
    node.prev = null;
    node.next = null;
    this.length--;
};

LinkedList.Node = function(data) {
    this.prev = null;
    this.next = null;
    this.data = data;
};
	

/**************************************
* BlipControl
***************************************/

function BlipControl(apiClient, swfUrl, containerDiv, options) {
    this.apiClient = apiClient;
    this._swfUrl = swfUrl;
    this._containerDiv = containerDiv;
    
    this.options = options || {}
    
    // Create custom events (requires YUI)
    this.events = {}
    
    this.events.init = new YAHOO.util.CustomEvent("init", this);
    this.events.ping = new YAHOO.util.CustomEvent("ping", this);
    
    this.events.preview = new YAHOO.util.CustomEvent("preview", this);
    this.events.previewPlaying = new YAHOO.util.CustomEvent("previewPlaying", this);
    this.events.previewStop = new YAHOO.util.CustomEvent("previewStop", this);

    this.events.create = new YAHOO.util.CustomEvent("create", this);
    this.events.add = new YAHOO.util.CustomEvent("add", this);
    this.events.remove = new YAHOO.util.CustomEvent("remove", this);

    this.events.play = new YAHOO.util.CustomEvent("play", this);
    this.events.pause = new YAHOO.util.CustomEvent("pause", this);
    this.events.stop = new YAHOO.util.CustomEvent("stop", this);
    this.events.next = new YAHOO.util.CustomEvent("next", this);
    this.events.prev = new YAHOO.util.CustomEvent("prev", this);
    
    this.events.volume = new YAHOO.util.CustomEvent("volume", this);
    
    this.events.playing = new YAHOO.util.CustomEvent("playing", this);
    this.events.buffering = new YAHOO.util.CustomEvent("buffering", this);
    this.events.progress = new YAHOO.util.CustomEvent("progress", this);
    this.events.loadComplete = new YAHOO.util.CustomEvent("loadComplete", this);
    this.events.playTime = new YAHOO.util.CustomEvent("playTime", this);
    this.events.trackListen = new YAHOO.util.CustomEvent("trackListen", this);
    this.events.nowPlaying = new YAHOO.util.CustomEvent("nowPlaying", this);
    
    this.events.ioError = new YAHOO.util.CustomEvent("ioError", this);
    this.events.previewIOError = new YAHOO.util.CustomEvent("previewIOError", this);
    
    // Crete key listeners
    this.keylisteners = {};
    
    this.keylisteners.shortcuts = new YAHOO.util.KeyListener(document, { keys: [32, 37, 39, 45, 61, 80, 107, 109, 187, 189] }, {
        fn: this.doKeyPress,
        scope: this,
        correctScope: true
    });
}
BlipControl.prototype = {
    _player: null,    
    _direction: 1,
    _volume: 1,
    _ping: 45000,
    _timestamp: 0,
    _playlistMax: 0,
    _playlistType: 'public',
    
    _init: false,
    _requests: new Array(),
    _playlist: new LinkedList.Circular(),
    _nodesById: new Array(),
    _currentNode: null,
    _currentPlayStart: 0,
    _currentPlayTime: 0,
    _currentTrackLength: 0,
    _preview: null,
    _playing: false,
    
    loadPlayer: function(swfUrl, containerDiv) {
        var swfUrl = swfUrl || this._swfUrl;
        var containerDiv = containerDiv || this._containerDiv;
    
        // TODO: Move this
        var that = this
        doneInit = function () {
            console.log("doneInit: BlipPlayer loaded");
            that.init();
        }

        this._playerId = "blipPlayer";

        // Load the player swf
        var flashvars = {};
        var attributes = {
            id: this._playerId
        };
        var params = {
            menu: "false",
            scale: "noscale",
            wmode: "transparent",
            allowfullscreen: "false",
            allowscriptaccess: "sameDomain"
        };
        try {
            swfobject.embedSWF(swfUrl, containerDiv, "5", "5", "9.0.0", false, flashvars, params, attributes);
        } catch (e) {
            console.log(e);
        }
    },
    
    init: function() {
        console.log('Init');
        
        var options = this.options;
        
        // Set options
        this._player = $(this._playerId);
        this._direction = options.direction || -1;
        this._volume = (options.volume != 'undefined')? options.volume : 1;
        this._ping = (options.ping != 'undefined')? options.ping * 1000 : 45000;
        this._timestamp = options.timestamp || new Date().getTime();
        this._playlistMax = options.playlistMax || 0;
        this._playlistType = options.playlistType || 'public';
        
        // Set the volume
        this._player.setVolume(this._volume);

        // Enable key listeners
        this.keylisteners.shortcuts.enable();
        
        // Prevent browser from going nuts when unloading the page
        $E.addListener(window, 'beforeunload', this.shutdown, this, this);
        
        // Finish init
        this._init = true;
        this.events.init.fire();
    },
    
    shutdown: function () {
        console.log('Shutdown');
        
        this.stop();
        this._init = false;
    },
    
    setCredentials: function (username, password) {
        this.apiClient.setCredentials(username, password);
    },

    load: function (a) {
        console.log("Load");
        
        for(var i = 0; i < a.length; i++) {
            this.add(a[i].id, a[i].url, a[i].blipType);
        }
    },
    
    create: function (o) {
        console.log("Create", o);
        
        this.events.create.fire(o);
        
        this.add(o.id, o.url, o.type);
    },
    
    add: function(id, url, type) {
        console.log("Add", id, url);
        
        // TODO: Make better
        var type = type || 'songUrl'; 
        
        if(!this._init) return false;
    
        var node = new LinkedList.Node({id: id, url: url, available: true})
        this._playlist.append(node);
        this._nodesById[id] = node;
        
        // Add song to blip player
        this._player.addSong({id: id, blipType: type, data: {path: url}});
        
        if(this._playlistMax && (this._playlist.length > this._playlistMax)) {
            this.remove();
        }
        
        this.events.add.fire({id: id, url: url, available: true});
    },
    
    remove: function(id) {
        console.log("Remove", id);
        
        var node = (id)? this._nodesById[id] : this._playlist.first;
        this._playlist.remove(node);
        
        //if(this._currentNode && (this._currentNode.data.id == node.data.id)) this._currentNode = null;
        
        this.events.remove.fire(node.data);
    },
    
    inheritPreview: function(id) {
        console.log("Inherit Preview", id);
        
        if(!this._init) return false;
        
        var prevNode = this._currentNode;
        this._currentNode = this._nodesById[id];
        
        this.events.play.fire(this._currentNode.data, (prevNode)? prevNode.data : null);

        this._currentPlayStart = this.getUTCTime();
        this._currentPlayTime = 0;
        this._currentTrackLength = 0;
        this._player.inheritPreview ();
        
        this.events.nowPlaying.fire(this._currentNode.data);
        
        this._playing = true;
        
        // TODO: HACK!!
        startedPLaying();
    },
    
    play: function(id) {
        console.log("Play", id);
        
        if(!this._init) return false;
        
        var id = id || null;
        var prevNode = null;
                
        if(! (id == null && this._currentNode)) {
            var prevNode = this._currentNode;
            
            if(id !== null) {
                if(this._nodesById[id] && !this._nodesById[id].data.available) return false;
                this._currentNode = this._nodesById[id];
            } else {
                if(this._direction >= 0) {
                    this._currentNode = this._playlist.first;
                } else {
                    this._currentNode = this._playlist.last;
                }
                id = this._currentNode.data.id;
            }
        }

        // Stop the preview
        if(this._preview) this.previewStop();
        
        this.events.play.fire(this._currentNode.data, (prevNode)? prevNode.data : null);

        if(id == null && this._currentNode) {
            this._player.playBlip();
        } else {
            this._currentPlayStart = this.getUTCTime();
            this._currentPlayTime = 0;
            this._currentTrackLength = 0;
            this._player.playByBlipId(this._currentNode.data.id);
        }
        
        this._playing = true;
    },
    
    pause: function() {
        console.log("Pause");
        
        if(!this._currentNode) return this.stop();
        
        this._player.pause();
        
        this.events.pause.fire(this._currentNode.data);
        
        this._playing = false;
    },
    
    stop: function() {
        console.log('Stop');
        
        if(this._currentNode) this.trackListen();
        
        this._player.stopBlip();
        
        this.events.stop.fire((this._currentNode)? this._currentNode.data : null);
        
        this._currentNode = null;
        this._playing = false;
    },
    
    next: function() {
        console.log("Next");
        
        if(!this._init) return false;
        this.trackListen();
        
        var node = this._currentNode;
        do {
            var prev = node;
            
            if(this._direction >= 0) {
                node = (node && node.next)? node.next : this._playlist.first;
            } else {
                node = (node && node.prev)? node.prev : this._playlist.last;
            }
            
            // Stop playing if only one song and it's unavailable
            //if(prev && node.data.id == prev.data.id && node.data.available == false) {
            if(prev && node.data.id == prev.data.id) {
                return this.stop();
            }
        } while(node.data.available == false);
        
        this.events.next.fire(node.data);
        
        this.play(node.data.id);
    },
    
    prev: function() {
        console.log("Prev");
        
        if(!this._init) return false;
        this.trackListen();
        
        var node = this._currentNode;
        do {
            if(this._direction >= 0) {
               node = (node)? node.prev : this._playlist.last;
            } else {
                node = (node)? node.next : this._playlist.first;
            }
        } while(node.data.available == false);
        
        this.events.prev.fire(node.data);
        this.play(node.data.id);
    },
    
    preview: function(url, blipType) {
        console.log("Preview");
        
        var blipType = blipType || 'songUrl';
        
        this.pause();
        
        if(this._preview) this.previewStop();
        
        this._player.preview(url, blipType);
        this._preview = true;
        
        this.events.preview.fire();
    },
    
    previewStop: function() {
        console.log("Stop preview");
         
        this._player.previewStop();
        this._preview = false;
        
        this.events.previewStop.fire();
    },
    
    previewError: function() {
        console.log("Preview error");
        
        this._preview = false;
        
        BlipControl.events.previewIOError.fire();
    },
    
    setTrackLength: function(n) {
        this._currentTrackLength = n;
        
        this.events.loadComplete.fire(this._currentPlayTime || 0, this._currentTrackLength);
    },
    
    setCurrentPlayTime: function(n) {
        this._currentPlayTime = n;
        
        this.events.playTime.fire(this._currentPlayTime, this._currentTrackLength);
    },
    
    startPing: function(pingInterval) {
        var pingInterval = pingInterval || this._pingInterval || 60;
        var that = this;
        if(this._ping) this._pingInterval = setInterval (function () {that.ping() }, this._ping);
        else this.stopPing();
    },
    
    stopPing: function() {
        clearInterval(this._pingInterval);
        this._pingInterval = null;
    },
    
    ping: function() {
        console.log("Ping");
        
        if(this._ping && (!this._requests['ping'] || !this._requests['ping'].isCallInProgress())) {
            var timestamp = this._timestamp;
            this._requests['ping'] = this.apiClient.blip_ping(this._timestamp, this._playlistType, 0, this._playlistMax, {
                success: function (o) {
                    if(o.responseText) {
                        var response = YAHOO.lang.JSON.parse(o.responseText);
                        this._timestamp = response.status.requestTime;

                        if(response.result.collection.Blip.length) {
                            var blips = response.result.collection.Blip;
                            blips.reverse();

                            for(var i = 0; i < blips.length; i++) {
                                if(!this._nodesById[blips[i].id])
                                   this.create(blips[i]);
                            }
                        }
                    }
                },
                failure: function (o) {
                    console.log('failure', o);
                },
                scope: this
            });
        }
        
        this.events.ping.fire(this._pingUrl);
    },
    
    trackListen: function() {
        console.log("Track Listen");
        
        var node = this._currentNode;
        var trackLength = Math.round(this._currentTrackLength / 1000);
        var playTime = this._currentPlayTime;
        var playStart = this._currentPlayStart;
        
        if(!node) return false;
        
        this.events.trackListen.fire({data: node.data, trackLength: trackLength, playTime: playTime, playStart: playStart});
    },
    
    doKeyPress: function (e, o) {
        var key = o[0];
        var event = o[1];
        var target = $E.getTarget(event);
        
        if(target.tagName == 'INPUT' || target.tagName == 'TEXTAREA' || target.tagName == 'BUTTON' || event.metaKey || event.altKey)
            return false;
        
        switch(key) {
            case 32:
                (this._playing)? this.pause() : this.play();
                break;
            case 37:
                this.prev();
                break;
            case 39:
                this.next();
                break;
            case 80:
                this.ping();
                break;
            case 45:
            case 109:
            case 189:
                this.setVolume(-1);
                break;
            case 61:
            case 107:
            case 187:
                this.setVolume(1);
                break;
        }
        
        // Prevent bubbling
        $E.stopEvent(event);
    },
    
    setVolume: function (n) {
    
        var d = .05 * n;
        var v = this._volume + d;

        if(v > 1) {
            this._volume = 1;
        } else if(v < 0) {
            this._volume = 0;
        } else {
            this._volume = v;
        }
        
        console.log("Volume", this._volume);
        this._player.setVolume(this._volume);
        
        this.events.volume.fire(this._volume);
    },
    
    getUTCTime: function () {
        var date = new Date();
        return (date.getTime() + date.getTimezoneOffset());
    }
}


/**********************************************
* BlipPlayer.swf External Interface Callbacks *
**********************************************/

function doneInit() {
    // Overwrite Me
}

function Blip_IO_Error () {
    console.log("IO Error");
    
    // TODO: HACK!!
    BlipControl.events.ioError.fire(BlipControl._currentNode.data);
    
    BlipControl.next();
}

function Preview_IO_Error () {
    console.log("Preview IO Error");
    
    BlipControl.previewError();
}

function startedPLaying () {
    console.log ("Song started playing");
    
    // TODO: HACK!!
    BlipControl.events.playing.fire(BlipControl._currentNode.data);
}

function Preview_startedPLaying () {
    console.log ("Preview started playing");
    
    // TODO: HACK!!
    BlipControl.events.previewPlaying.fire(BlipControl._preview);
}

function notifyPercent (p) {
    // TODO: HACK!!
    BlipControl.events.progress.fire(p);
}

function notifyLength (n) {
    BlipControl.setCurrentPlayTime(n);
}

function doneTackLength (n) {
    if(!n) {
        Blip_IO_Error();
        return
    }
    
    BlipControl.setTrackLength(n);
}

function Blip_Next () {
    console.log("Next Song");
    BlipControl.next();
}

function playerMessage (str) {
    console.log ("BlipPlayer Message ", str);
}